import 'dart:async';
import 'dart:convert';

import 'package:flutter/material.dart';
import 'package:http/http.dart' as http;

class CodeChatUI<T> extends StatefulWidget {
  // url adalah endpoint dari chatbot
  final String url;

  // requestBuilder adalah fungsi yang mengambil input pengguna dan mengembalikan Map<String, dynamic>
  final Map<String, dynamic> Function(String input) requestBuilder;

  // onResponseSuccess adalah fungsi yang mengambil respons dan input pengguna
  final void Function(dynamic response, String input) onResponseSuccess;

  // onFailure adalah fungsi yang mengambil error, respons, dan input pengguna
  final void Function(dynamic error, dynamic response, String input)? onFailure;

  // chatsValueListenable adalah ValueNotifier yang berisi daftar percakapan
  final ValueNotifier<List<T>> chatsValueListenable;

  // chatBuilder adalah fungsi yang mengambil konteks dan percakapan, lalu mengembalikan widget
  final Widget Function(BuildContext context, T chat) chatBuilder;

  // appBar adalah widget yang akan ditampilkan sebagai app bar
  final PreferredSizeWidget? appBar;

  // sendButtonBuilder adalah fungsi yang mengambil konteks dan fungsi onSend, lalu mengembalikan widget
  final Widget Function(BuildContext context, void Function() onSend)?
      sendButtonBuilder;

  // construktor
  const CodeChatUI({
    Key? key,
    required this.url,
    required this.requestBuilder,
    required this.onResponseSuccess,
    required this.chatsValueListenable,
    required this.chatBuilder,
    this.onFailure,
    this.appBar,
    this.sendButtonBuilder,
  }) : super(key: key);

  @override
  State<CodeChatUI<T>> createState() => _CodeChatUIState<T>();
}

class _CodeChatUIState<T> extends State<CodeChatUI<T>> {
  final _controller = TextEditingController();
  final _scrollController = ScrollController();
  final _focusNode = FocusNode();
  final _loadingNotifier = ValueNotifier<bool>(false);

  @override
  void initState() {
    super.initState();
    if (_scrollController.hasClients) {
      _scrollController.animateTo(
        _scrollController.position.maxScrollExtent,
        duration: const Duration(seconds: 1),
        curve: Curves.fastOutSlowIn,
      );
    }
  }

  // onSend adalah fungsi yang akan dipanggil ketika tombol kirim ditekan
  // menggunakan package http untuk mengirim request ke endpoint
  Future<void> onSend() async {
    final input = _controller.text;
    if (input.isEmpty) return;
    _controller.clear();
    _loadingNotifier.value = true;
    try {
      final response = await http.post(
        Uri.parse(widget.url),
        headers: {'Content-Type': 'application/json'},
        body: json.encode(widget.requestBuilder(input)),
      );
      if (response.statusCode == 200) {
        widget.onResponseSuccess(json.decode(response.body), input);

        Timer(
          const Duration(milliseconds: 100),
          () => _scrollController
              .jumpTo(_scrollController.position.minScrollExtent),
        );
      } else {
        _controller.text = input;
        widget.onFailure?.call(
            'Error ${response.statusCode}', json.decode(response.body), input);
      }
      _loadingNotifier.value = false;
    } catch (e) {
      _controller.text = input;
      widget.onFailure?.call(e, null, input);
      _loadingNotifier.value = false;
    }
  }

  @override
  Widget build(BuildContext context) {
    return GestureDetector(
      onTap: () => _focusNode.unfocus(),
      child: Scaffold(
        appBar: widget.appBar,
        body: Column(
          children: [
            Expanded(
              child: ValueListenableBuilder<List<T>>(
                valueListenable: widget.chatsValueListenable,
                builder: (context, chats, child) => ListView.builder(
                  controller: _scrollController,
                  itemCount: chats.length,
                  reverse: true,
                  shrinkWrap: true,
                  padding: const EdgeInsets.symmetric(
                    horizontal: 15,
                    vertical: 15,
                  ),
                  itemBuilder: (context, index) {
                    return widget.chatBuilder(context, chats[index]);
                  },
                ),
              ),
            ),
            ValueListenableBuilder<bool>(
              valueListenable: _loadingNotifier,
              builder: (context, loading, child) => loading
                  ? const Center(
                      child: CircularProgressIndicator(),
                    )
                  : const SizedBox(),
            ),
            Padding(
              padding: const EdgeInsets.all(15),
              child: Row(
                children: [
                  Expanded(
                    child: TextFormField(
                      controller: _controller,
                      focusNode: _focusNode,
                      decoration: InputDecoration(
                        hintText: "Type a message",
                        border: OutlineInputBorder(
                          borderRadius: BorderRadius.circular(30),
                          borderSide:
                              const BorderSide(color: Colors.grey, width: 1),
                        ),
                        focusedBorder: OutlineInputBorder(
                          borderRadius: BorderRadius.circular(30),
                          borderSide:
                              const BorderSide(color: Colors.blue, width: 1),
                        ),
                        isDense: true,
                      ),
                    ),
                  ),
                  const SizedBox(width: 10),
                  widget.sendButtonBuilder?.call(context, () => onSend()) ??
                      IconButton(
                        onPressed: () => onSend(),
                        icon: const Icon(Icons.send),
                      ),
                ],
              ),
            ),
          ],
        ),
      ),
    );
  }
}
